__title__ = "rpcm"
__description__ = "Rational Polynomial Camera Model for optical satellite images."
__url__ = 'https://github.com/centreborelli/rpcm'
__version__ = "1.4.8"
__author__ = "Carlo de Franchis"
__author_email__ = 'carlo.de-franchis@ens-paris-saclay.fr'
